#
# Copyright (c) 2024-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of PyUtilities.
#
# PyUtilities is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyUtilities is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with PyUtilities. If not,
# see <https://www.gnu.org/licenses/>.
#

import pytz
import os
import sys
import shutil
import time
import tempfile
import errno
import stat
import glob
import platform
import subprocess
import re
import typing
import xml.etree.ElementTree as ET

if sys.platform != 'win32':
    import pty

from semver.version import Version
from datetime import datetime
from pathlib import Path
from types import TracebackType
from typing import List, Dict, Tuple, Callable, Any, Optional


# -- Private functions

def _handleRemoveReadonly(func: Callable[..., Any], path: str, exc: Tuple[type[BaseException], BaseException, TracebackType]) -> None:
    excvalue = typing.cast(OSError, exc[1])
    if func in (os.rmdir, os.remove, os.unlink) and excvalue.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # -- 0777
        func(path)
    else:
        raise


# -- Classes
class Utility:
    """Helper methods."""

    _app_name = ''
    _app_version = ''

    # -- This makes sure that some things work when running as an app and not from the command line.
    # -- For example this fixes the error: "The process has forked and you cannot use this CoreFoundation functionality safely. You MUST exec()."
    # -- when calling shellCommand().
    use_ui_application_mode = False

    # -- This is used in Unit tests to mock the time for 'now'.
    _mock_now_date: Optional[datetime] = None

    @classmethod
    def setAppInfo(cls, name: str, version: str) -> None:
        Utility._app_name = name
        Utility._app_version = name

    @classmethod
    def shellCommand(cls, command_and_args: List[str], from_dir: Optional[Path] = None, capture_output: bool = False, filter_ansi: bool = False) -> Tuple[int, List[str]]:
        try:
            captured_output: List[str] = []
            return_code = 0

            if sys.platform == 'win32' or Utility.use_ui_application_mode:
                if capture_output:
                    result = subprocess.run(command_and_args, cwd=from_dir, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
                    captured_output = result.stdout.decode('utf-8').split('\n')
                else:
                    result = subprocess.run(command_and_args, cwd=from_dir, stderr=sys.stderr, stdout=sys.stdout)

                return_code = result.returncode
            else:
                if from_dir is not None:
                    raise RuntimeError('from_dir is not supported when using pty.spawn.')

                if capture_output:
                    output_bytes = bytearray()

                    def read_pty_output(fd: int) -> bytes:
                        data = os.read(fd, 1024)

                        if len(data) != 0:
                            output_bytes.extend(data)

                            # -- We don't need to print anything out, we're just capturing.
                            data = bytearray()
                            data.append(0)

                        return data

                    return_code = pty.spawn(command_and_args, master_read=read_pty_output)
                    captured_output = output_bytes.decode('utf-8').split('\n')
                else:
                    return_code = pty.spawn(command_and_args)

                for i in range(len(captured_output)):
                    if captured_output[i].endswith('\r'):
                        captured_output[i] = captured_output[i].strip('\r')

            if filter_ansi:
                # -- If we are in UI mode then we filter any ANSI characters
                ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')

                new_output: List[str] = []
                for line in captured_output:
                    new_output.append(ansi_escape.sub('', line))

                captured_output = new_output

            return return_code, captured_output
        except Exception as e:
            raise RuntimeError('Error running shell command: ' + str(e))

    @classmethod
    def commandExists(cls, command: str) -> bool:
        return_code, captured_output = Utility.shellCommand(['where' if os.name == 'nt' else 'which', command], capture_output=True)
        return return_code == 0

    @classmethod
    def requireCommand(cls, command: str) -> None:
        if not Utility.commandExists(command):
            raise RuntimeError('Cannot find command "' + command + '".')

    @classmethod
    def stringToInt(cls, string: Optional[str]) -> Optional[int]:
        if string is None:
            return None

        return int(string)

    @classmethod
    def stringToFloat(cls, string: Optional[str]) -> Optional[float]:
        if string is None:
            return None

        return float(string)

    @classmethod
    def dateFromString(cls, string: Optional[str], format: str, utc: bool = False) -> Optional[datetime]:
        if string is None:
            return None

        try:
            date = datetime.strptime(string, format)
            if utc:
                date = pytz.utc.localize(date)

            return date
        except ValueError:
            return None

    @classmethod
    def utcTimeNow(cls) -> datetime:
        if Utility._mock_now_date is not None:
            return Utility._mock_now_date

        return datetime.now().astimezone(pytz.utc)

    @classmethod
    def utcDatetime(cls, year: int, month: int, day: int, hour: int, minutes: int, seconds: int) -> datetime:
        result: datetime = pytz.utc.localize(datetime(year, month, day, hour, minutes, seconds))
        return result

    @classmethod
    def xmlElementToString(cls, element: ET.Element, xml_declaration: bool = False) -> str:
        return ET.tostring(element, encoding='unicode', short_empty_elements=False, xml_declaration=xml_declaration)

    @classmethod
    def processIsRunning(cls, process_name: str) -> bool:
        return_code, captured_output = Utility.shellCommand(['ps', '-axc', '-o', 'comm'])
        return process_name in captured_output

    @classmethod
    def lookInFolderFor(cls, folder: Path, wildcard: str) -> List[str]:
        # -- We use this here instead of just simply Path.exists()
        # -- because we want the test to be case-sensitive on all platforms,
        # -- so we list what the match are and let glob give us the paths.
        paths_found = []
        looking_in = folder / wildcard
        prefix = str(folder / ' ')[:-1]

        for p in glob.glob(str(looking_in), recursive=True):
            as_string = str(p)
            if len(as_string) > 4:
                if len(prefix) != 0 and as_string.startswith(prefix):
                    as_string = as_string[len(prefix):]

                paths_found.append(as_string)

        return paths_found

    @classmethod
    def deleteFolder(cls, folder: Path, force_delete: bool = False) -> None:
        if folder.exists():
            if force_delete:
                ignore_errors = False
                on_error = _handleRemoveReadonly
            else:
                ignore_errors = True
                on_error = None

            shutil.rmtree(folder, ignore_errors=ignore_errors, onerror=on_error)

    @classmethod
    def fileOlderThan(cls, path: Path, time_in_seconds: int) -> bool:
        if not path.exists():
            return True

        return (time.time() - path.stat().st_mtime) > time_in_seconds

    @classmethod
    def appFolder(cls, create_if_needed: bool = False) -> Path:
        if Utility._app_name == '':
            raise RuntimeError('Utility.setAppInfo() needs to be called during app init.')

        if platform.system() != 'Darwin':
            # TODO: Paths need to be ported to 'Windows' and 'Linux'
            raise RuntimeError('Utility.appFolder() is macOS only at the moment.')

        folder = Path.home() / 'Library' / 'Application Support' / f'net.malenfant.{Utility._app_name}'
        if create_if_needed:
            folder.mkdir(parents=True, exist_ok=True)

        return folder

    @classmethod
    def tempFolder(cls, create_if_needed: bool = False) -> Path:
        if Utility._app_name == '':
            raise RuntimeError('Utility.setAppInfo() needs to be called during app init.')

        folder = Path(tempfile.gettempdir()) / f'net.malenfant.{Utility._app_name}' / str(os.getpid())
        if create_if_needed:
            folder.mkdir(parents=True, exist_ok=True)

        return folder

    @classmethod
    def deleteTempFolder(cls) -> None:
        folder = Utility.tempFolder()
        if folder.exists():
            Utility.deleteFolder(folder, force_delete=True)

    @classmethod
    def checkForUpdates(cls, force_check: bool) -> None:
        try:
            if Utility._app_name == '' or Utility._app_version == '':
                return

            out_file = Utility.appFolder(create_if_needed=True) / 'app-update-check'
            if not force_check and not Utility.fileOlderThan(out_file, time_in_seconds=(24 * 60 * 60)):
                return

            latest_version = Git(f'codeberg.org/DidierMalenfant/{Utility._app_name}').getLatestVersion()
            if latest_version is None:
                return

            if out_file.exists():
                out_file.unlink()

            out_file.write_text('check')

            if latest_version > Version.parse(Utility._app_version):
                warning = '‼️' if sys.platform == 'darwin' else '!!'
                print(f'{warning}  Version v{str(latest_version)} is available for Main. You have v{Utility._app_version} {warning}')
                print(f'Please run "pip install {Utility._app_name} --upgrade" to upgrade.')
        except Exception:
            pass


class Git:
    """Utility methods for git repos."""

    def __init__(self, url: str, url_is_local_folder: bool = False):
        """Setup access to the git repo at url."""

        if not Utility.commandExists('git'):
            raise RuntimeError('You must have git installed on your machine to continue.')

        self._url = url if url_is_local_folder else 'https://' + url + '.git'
        self._refs: Optional[Dict[str, str]] = None
        self._tags: Optional[List[str]] = None
        self._tag_versions: Optional[List[Version]] = None
        self._branches: Optional[Dict[str, str]] = None
        self._head_branch: Optional[str] = None
        self._latest_version: Optional[Version] = None

    def _cmd(self, arguments: List[str], folder: Optional[Path] = None) -> List[str]:
        arguments.insert(0, 'git')
        arguments.append(self._url.replace('https://', 'https://anonymous:@'))

        if folder is not None:
            arguments.append(str(folder))

        return_code, captured_output = Utility.shellCommand(arguments, capture_output=True)
        if return_code != 0:
            error = captured_output[0]

            if error.startswith('usage: git'):
                # -- git is giving us the usage info back it seems.
                raise SyntaxError('Invalid git command line')
            elif error.startswith('fatal: not a git repository (or any of the parent directories): .git'):
                raise RuntimeError('Your project folder needs to be a git repo for certain commands to work correctly. Try `git init` to create one.')
            elif error == 'remote: Invalid username or password.':
                raise RuntimeError('Cannot access git repo at "' + self._url + '". Maybe it is private?')
            else:
                # -- Or maybe something else went wrong.
                raise RuntimeError('Error running git: ' + error)

        return captured_output

    def listRefs(self) -> Dict[str, str]:
        if self._refs is None:
            self._refs = {}
            for ref in self._cmd(['ls-remote', '--refs']):
                refs_index = ref.find('refs/')
                if refs_index >= 0:
                    self._refs[ref[refs_index + 5:]] = ref[:40]

        return self._refs

    def listBranches(self) -> Dict[str, str]:
        if self._branches is None:
            self._branches = {}
            refs = self.listRefs()
            for ref in refs.keys():
                if ref.startswith('heads/'):
                    self._branches[ref[6:]] = refs[ref]

        return self._branches

    def getHeadBranch(self) -> str:
        if self._head_branch is None:
            for line in self._cmd(['remote', 'show']):
                if line.startswith('  HEAD branch:'):
                    self._head_branch = line[15:]

            if self._head_branch is None:
                raise RuntimeError('Cannot find head branch for "' + self._url + '".')

        return self._head_branch

    def listTags(self) -> List[str]:
        if self._tags is None:
            self._tags = []
            for ref in self.listRefs().keys():
                if ref.startswith('tags/'):
                    tag = ref[5:]
                    if not tag.startswith('@'):
                        self._tags.append(tag)

        return self._tags

    def listTagVersions(self) -> List[Version]:
        if self._tag_versions is None:
            self._tag_versions = []

            for tag in self.listTags():
                try:
                    if tag.startswith('v'):
                        tag = tag[1:]

                    self._tag_versions.append(Version.parse(tag))
                except ValueError:
                    pass

            self._tag_versions = sorted(self._tag_versions)

        return self._tag_versions

    def getLatestVersion(self) -> Optional[Version]:
        if self._latest_version is None:
            all_versions = self.listTagVersions()

            if len(all_versions) > 0:
                self._latest_version = all_versions[-1]

        return self._latest_version

    def getLatestCommitHashForBranch(self, branch_name: str) -> Optional[str]:
        return self.listBranches().get(branch_name)

    def isABranch(self, name: str) -> bool:
        return name in self.listBranches()

    def isATag(self, name: str) -> bool:
        for tag in self.listTags():
            if tag == name:
                return True

        return False

    def cloneIn(self, folder: Path, branch: Optional[str] = None, recurse_submodules: bool = True) -> None:
        folder.mkdir(parents=True, exist_ok=True)

        command_line: List[str] = ['clone', '--quiet', '--depth', '1']

        if recurse_submodules:
            command_line.append('--recurse-submodules')

        if branch is not None:
            command_line.append('--branch')
            command_line.append(branch)

        self._cmd(command_line, folder)



#
# Copyright (c) 2025-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of p-asm.
#
# p-asm is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# p-asm is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with p-asm. If not,
# see <https://www.gnu.org/licenses/>.
#

import getopt
import sys
import traceback

from typing import List, Dict, Callable
from pathlib import Path

from .__about__ import __appname__
from .__about__ import __version__
from .Exceptions import ArgumentError

# -- We need to import from our parent folder here.
sys.path.append(str(Path(sys.path[0]) / '..'))

from PyUtilities.Utility import Utility     # noqa: E402

_verbose_mode = False


def printUsage(commands: List[str]) -> None:
    if len(commands) > 1:
        switch: Dict[str, Callable[[List[str]], None]] = {
            'topics': printTopics,
            'license': printLicense
        }

        method = switch.get(commands[1])
        if method is None:
            raise ArgumentError('Unknown topic "' + commands[1] + '".')

        method(commands)
        return

    printVersion(commands)
    print('')
    print(f'usage: {__appname__} <options> <command> <arguments> <path>')
    print('')
    print('The following commands are supported:')
    print('')
    print('   help <topic>       - Show a help message. topic is optional (use "help topics" for a list).')
    print('   version            - Print the current version.')
    print('')
    print('The following options are supported:')
    print('')
    print('   --debug/-d         - Enable extra debugging information.')
    print('   --verbose/-v       - Enable verbose mode (prints information on the tracks affected).')
    print('')
    print(f'{__appname__} is free software, type "{__appname__} help license" for license information.')


def printVersion(commands: List[str]) -> None:
    print(f'💾 {__appname__} v{__version__} 💾')


def printTopics(commands: List[str]) -> None:
    printVersion(commands)
    print('')
    print('Usage:')
    print(f'   {__appname__} help license - Show the license for the app.')
    print('')


def printLicense(commands: List[str]) -> None:
    printVersion(commands)
    print('')
    print('GPL License Version 3')
    print('')
    print('Copyright (c) 2025-present Didier Malenfant <didier@malenfant.net>')
    print('')
    print(f'{__appname__} is free software: you can redistribute it and/or modify it under the terms of the GNU General')
    print('Public License as published by the Free Software Foundation, either version 3 of the License, or')
    print('(at your option) any later version.')
    print('')
    print(f'{__appname__} is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the')
    print('implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public')
    print('License for more details.')
    print('')
    print(f'You should have received a copy of the GNU General Public License along with {__appname__}. If not,')
    print('see <https://www.gnu.org/licenses/>.')
    print('')


def main() -> None:
    global _verbose_mode

    debug_on = False

    Utility.setAppInfo(__appname__, __version__)

    try:
        # -- Gather the arguments, remove the first argument (which is the script filename)
        opts, commands = getopt.getopt(sys.argv[1:], 'tdv', ['help', 'test', 'debug', 'verbose', 'only='])

        for o, a in opts:
            if o in ('-d', '--debug'):
                print('Enabling debugging information.')
                debug_on = True
            elif o in ('-v', '--verbose'):
                print('Enabling verbose mode.')
                _verbose_mode = True
            elif o in ('--help'):
                commands = ['help']

        if len(commands) == 0:
            raise ArgumentError(f'Expected a command! Maybe start with `{__appname__} help`?')

        switch: Dict[str, Callable[[List[str]], None]] = {
            'help': printUsage,
            'version': printVersion
        }

        if commands is None:
            raise ArgumentError(f'Expected a command! Maybe start with `{__appname__} help`?')

        command = commands[0]
        method = switch.get(command)
        if method is None:
            raise ArgumentError('Unknown commanwd "' + command + '".')

        method(commands)

    except getopt.GetoptError:
        printUsage([])
    except Exception as e:
        if debug_on:
            print(traceback.format_exc())
        else:
            print(f'Error: {e}')

        sys.exit(1)
    except KeyboardInterrupt:
        print('Execution interrupted by user.')
        pass


if __name__ == '__main__':
    main()



#
# Copyright (c) 2024-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of PyUtilities.
#
# PyUtilities is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyUtilities is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with PyUtilities. If not,
# see <https://www.gnu.org/licenses/>.
#

import sys
import getopt
import traceback

from os.path import normpath
from pathlib import Path, PureWindowsPath
from typing import List, Dict, Callable

# -- We need to import from our parent folder here.
sys.path.append(str(Path(sys.path[0]) / '..'))

from PyUtilities.Utility import Utility     # noqa: E402
from PyUtilities.Utility import Git         # noqa: E402

__appname__ = 'DidierCI'
__version__ = '0.9'
_head_run = False
_verbose_on = False


def printVersion(commands: List[str]) -> None:
    print(f'🌡️   {__appname__} v{__version__} 🌡️')


def printUsage(commands: List[str]) -> None:
    if len(commands) > 1:
        switch: Dict[str, Callable[[List[str]], None]] = {
            'topics': printTopics,
            'license': printLicense,
            'run': printHelpRun,
            'install': printHelpInstall
        }

        method = switch.get(commands[1])
        if method is None:
            raise RuntimeError('Unknown topic "' + commands[1] + '".')

        method(commands)
        return

    printVersion(commands)
    print('')
    print('usage: DidierCI.py <options> commands')
    print('')
    print('The following commands are supported:')
    print('')
    print('   help <topic>    - Show a help message. topic is optional (use "help topics" for a list).')
    print('   version         - Print the current version.')
    print('   run tasks       - Run the given tasks on the local repo.')
    print('   install tasks   - Install tasks to be run pre and post commit on the local repo.')
    print('')
    print('The following options are supported:')
    print('')
    print('   --debug/-d     - Enable extra debugging information.')
    print('   --verbose/-v   - Print tasks output if any.')
    print('')
    print('DidierCI is free software, run "DidierCI help license" for license information.')


def printTopics(commands: List[str]) -> None:
    printVersion(commands)
    print('')
    print('Usage:')
    print('   DidierCI.py help license - Show the license for the app.')
    print('   DidierCI.py help run     - Show help about the run command.')
    print('   DidierCI.py help install - Show help about the install command.')
    print('')


def printLicense(commands: List[str]) -> None:
    printVersion(commands)
    print('')
    print('GPL License Version 3')
    print('')
    print('Copyright (c) 2024-present Didier Malenfant <didier@malenfant.net>')
    print('')
    print('DidierCI is free software: you can redistribute it and/or modify it under the terms of the GNU General')
    print('Public License as published by the Free Software Foundation, either version 3 of the License, or')
    print('(at your option) any later version.')
    print('')
    print('DidierCI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the')
    print('implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public')
    print('License for more details.')
    print('')
    print('You should have received a copy of the GNU General Public License along with Main. If not,')
    print('see <https://www.gnu.org/licenses/>.')
    print('')


def printHelpRun(commands: List[str]) -> None:
    printVersion(commands)
    print('')
    print('Usage:')
    print('   DidierCI.py <options> run tasks  - Run the given tasks on the local repo.')
    print('')
    print('Run tasks on the local repository. The following tasks are supported:')
    print('')
    print('   flake8            - Run flake 8 linting on any folder which contains python files.')
    print('   mypy              - Run mypy to ensure correct typing information on any folder which contains python files.')
    print('   pytest            - Run pytest is the root folder containt a folder named Tests.')
    print('')
    print('The following options are supported:')
    print('')
    print('   --head         - Run the commands on the current repo\'s head commit.')
    print('')
    print('You can also list more than one task on the command line.')


def printHelpInstall(commands: List[str]) -> None:
    printVersion(commands)
    print('')
    print('Usage:')
    print('   DidierCI.py install tasks - Install tasks to be run pre and post commit on the local repo.')
    print('')
    print('Install tasks to run on the local repository during pre and post git commits. The following tasks are supported:')
    print('')
    print('   flake8            - Run flake 8 linting on any folder which contains python files.')
    print('   mypy              - Run mypy to ensure correct typing information on any folder which contains python files.')
    print('   pytest            - Run pytest is the root folder containt a folder named Tests.')
    print('')
    print('You can also list more than one task on the command line.')


def findFoldersWithPythonFiles(in_folder: Path) -> List[Path]:
    found_folders: List[Path] = []

    if len(Utility.lookInFolderFor(in_folder, '*.py')) != 0:
        found_folders.append(in_folder)

    for item in Utility.lookInFolderFor(in_folder, '*'):
        item_found = in_folder / item

        if not item_found.is_dir():
            continue

        found_folders += findFoldersWithPythonFiles(item_found)

    return found_folders


def findSubModulesFolders(in_folder: Path) -> List[Path]:
    found_folders: List[Path] = []

    maybe_hooks_folder = in_folder / 'hooks'

    if maybe_hooks_folder.exists() and maybe_hooks_folder.is_dir():
        found_folders.append(in_folder)
    else:
        for item in Utility.lookInFolderFor(in_folder, '*'):
            item_found = in_folder / item

            if not item_found.is_dir():
                continue

            found_folders += findSubModulesFolders(item_found)

    return found_folders


def runMyPy(in_folder: Path) -> None:
    if not Utility.commandExists('mypy'):
        print('WARNING: Unable to find mypy for CI testing.')
        return

    global _verbose_on

    for folder in findFoldersWithPythonFiles(in_folder):
        return_code, captured_output = Utility.shellCommand(['mypy', '--disallow-any-generics', '--ignore-missing-imports', '--follow-imports=silent',
                                                             '--no-incremental', '--strict-equality', '--disallow-incomplete-defs',
                                                             '--disallow-redefinition', '--disallow-untyped-globals', '--no-implicit-optional',
                                                             '--no-implicit-reexport', '--warn-redundant-casts', '--warn-unused-ignores',
                                                             '--warn-unreachable', '--warn-no-return', '--disallow-untyped-calls',
                                                             '--disallow-untyped-defs', '--check-untyped-defs', '--disallow-any-generics',
                                                             '--warn-return-any', str(folder)], capture_output=not _verbose_on, filter_ansi=not _verbose_on)

        if return_code != 0:
            if not _verbose_on:
                for line in captured_output:
                    if len(line) > 0:
                        print(line)

            raise RuntimeError()


def runFlake8(in_folder: Path) -> None:
    if not Utility.commandExists('flake8'):
        print('WARNING: Unable to find flake8 for CI testing.')
        return

    global _verbose_on

    for folder in findFoldersWithPythonFiles(in_folder):
        return_code, captured_output = Utility.shellCommand(['flake8', str(folder)], capture_output=not _verbose_on, filter_ansi=not _verbose_on)

        if return_code != 0:
            if not _verbose_on:
                for line in captured_output:
                    if len(line) > 0:
                        print(line)

            raise RuntimeError()


def runPyTest(in_folder: Path) -> None:
    if not Utility.commandExists('pytest'):
        print('WARNING: Unable to find pytest for CI testing.')
        return

    global _verbose_on

    tests_folder = in_folder / 'Tests'
    if tests_folder.exists():
        return_code, captured_output = Utility.shellCommand(['pytest', str(tests_folder)], capture_output=not _verbose_on, filter_ansi=not _verbose_on)

        if return_code != 0:
            if not _verbose_on:
                for line in captured_output:
                    if len(line) > 0:
                        print(line)

            raise RuntimeError()


def runCI(commands: List[str]) -> None:
    global _head_run

    switch: Dict[str, Callable[[Path], None]] = {
        'flake8': runFlake8,
        'mypy': runMyPy,
        'pytest': runPyTest
    }

    try:
        in_folder = Path('.')
        if _head_run:
            in_folder = Utility.tempFolder(create_if_needed=True)

            git_repo = Git('.', url_is_local_folder=True)
            git_repo.cloneIn(in_folder)

        for task in commands[1:]:
            method = switch.get(task)
            if method is None:
                raise RuntimeError('Unknown task "' + task + '".')

            method(in_folder)
    finally:
        Utility.deleteTempFolder()


def installCI(commands: List[str]) -> None:
    local_folder = Path('.') / '.git'
    repo_folders = [local_folder]

    repo_folders += findSubModulesFolders(local_folder / 'modules')

    for folder in repo_folders:
        if not folder.exists() or not folder.is_dir():
            raise RuntimeError(f'Folder "{folder}"has no git repo to install hooks.')

        valid_tasks: List[str] = ['flake8', 'mypy', 'pytest']
        tasks_string = ''
        for task in commands[1:]:
            if task not in valid_tasks:
                raise RuntimeError('Unknown task "' + task + '".')

            if len(tasks_string) != 0:
                tasks_string += ' '

            tasks_string += task

        path_to_this_file = Path(sys.path[0]).relative_to(Path('.').resolve()) / 'DidierCI.py'
        path_to_this_file_as_posix = PureWindowsPath(normpath(PureWindowsPath(path_to_this_file).as_posix())).as_posix()

        with open(folder / 'hooks' / 'post-commit', 'w') as out_file:
            out_file.write('#!/bin/bash\n')
            out_file.write('#\n')
            out_file.write('# post-commit git-hook for DidierCI.\n')
            out_file.write('#\n')
            out_file.write('# This file is auto-generated. DO NOT EDIT.\n')
            out_file.write('#\n')
            out_file.write('\n')
            out_file.write('unset GIT_DIR\n')
            out_file.write('unset GIT_WORK_TREE\n')
            out_file.write('\n')
            out_file.write('OPERATION_FAILED=0\n')
            out_file.write(f'python {path_to_this_file_as_posix} --head run {tasks_string} || OPERATION_FAILED=1\n')
            out_file.write('\n')
            out_file.write('if [[ $OPERATION_FAILED -eq 0 ]]\n')
            out_file.write('then\n')
            out_file.write('   exit\n')
            out_file.write('fi\n')
            out_file.write('\n')
            out_file.write('exec 5>&1\n')
            out_file.write(f'TEST=$(python {path_to_this_file_as_posix} --head run {tasks_string})\n')
            out_file.write('\n')
            out_file.write('if [[ "$OSTYPE" != "darwin"* ]]; then\n')
            out_file.write('   echo $TEST\n')
            out_file.write('   exit 1\n')
            out_file.write('fi\n')
            out_file.write('\n')
            out_file.write('/usr/bin/osascript <<-EOF\n')
            out_file.write('\n')
            out_file.write(' tell application "System Events"\n')
            out_file.write('      activate\n')
            out_file.write('      display alert "Last commit failed CI tests." message "$TEST" as critical\n')
            out_file.write(' end tell\n')
            out_file.write('\n')
            out_file.write('EOF\n')

            out_file.close()

        with open(folder / 'hooks' / 'pre-commit', 'w') as out_file:
            out_file.write('#!/bin/bash\n')
            out_file.write('#\n')
            out_file.write('# pre-commit git-hook for DidierCI.\n')
            out_file.write('#\n')
            out_file.write('# This file is auto-generated. DO NOT EDIT.\n')
            out_file.write('#\n')
            out_file.write('\n')
            out_file.write('# -- Stash any unstaged changes\n')
            out_file.write('git stash -q --keep-index\n')
            out_file.write('\n')
            out_file.write('# -- Run the pre-commit tests\n')
            out_file.write(f'python {path_to_this_file_as_posix} run {tasks_string}\n')
            out_file.write('\n')
            out_file.write('# -- Store the last exit code in a variable\n')
            out_file.write('RESULT=$?\n')
            out_file.write('\n')
            out_file.write('# -- Unstash the stashed changes\n')
            out_file.write('git stash pop -q\n')
            out_file.write('\n')
            out_file.write('# -- Return the exit code\n')
            out_file.write('exit $RESULT\n')
            out_file.write('\n')
            out_file.write('# << must have a newline after the above command >>\n')
            out_file.write('\n')

            out_file.close()


def main() -> None:
    global _head_run
    global _verbose_on

    _debug_on = False

    Utility.setAppInfo(__appname__, __version__)

    try:
        # -- Gather the arguments, remove the first argument (which is the script filename)
        opts, commands = getopt.getopt(sys.argv[1:], 'dhv', ['help', 'debug', 'head', 'verbose'])

        for o, a in opts:
            if o in ('-d', '--debug'):
                print('Enabling debugging information.')
                _debug_on = True
            elif o in ('--help'):
                commands = ['help']
            elif o in ('-h', '--head'):
                _head_run = True
            elif o in ('-v', '--verbose'):
                _verbose_on = True

        if commands is None or len(commands) == 0:
            raise RuntimeError('Expected a command! Maybe start with `DidierCI.py help`?')

        switch: Dict[str, Callable[[List[str]], None]] = {
            'help': printUsage,
            'version': printVersion,
            'run': runCI,
            'install': installCI
        }

        command: str = commands[0]
        method = switch.get(command)
        if method is None:
            raise RuntimeError('Unknown command "' + command + '".')

        method(commands)

    except getopt.GetoptError:
        printUsage([])
    except Exception as e:
        if _debug_on:
            print(traceback.format_exc())
        else:
            exception_string = str(e)
            if len(exception_string) != 0:
                print(f'Error: {e}')

        sys.exit(1)
    except KeyboardInterrupt:
        print('Execution interrupted by user.')
        pass


if __name__ == '__main__':
    main()
